/***************************************************************************
 *cr
 *cr            (C) Copyright 2007 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/
/*
 * CUDA accelerated coulombic potential grid test code
 *   John E. Stone <johns@ks.uiuc.edu>
 *   http://www.ks.uiuc.edu/~johns/
 *
 * Coulombic potential grid calculation microbenchmark based on the time
 * consuming portions of the 'cionize' ion placement tool.
 *
 * This version of the code uses the 64KB constant buffer area reloaded 
 * for each group of MAXATOMS atoms, until the contributions from all 
 * atoms have been summed into the potential grid.  This GPU code currenty
 * outruns one core of the 2.6GHz Intel Core 2 Quad host CPU by a factor
 * of 37:1 versus our best performing hand-optimized C version with 
 * fully vectorized loops and SSE via the Intel 9.0 compilers.
 * The performance difference versus one CPU of the SGI Altix (Itanium2)
 * is even greater, at a ratio of 54:1.
 *
 * The full multi-GPU implementation (one thread per GPU) in a 
 * GPU-accelerated test version of 'cionize' achieves the same performance
 * as 160 processors of the SGI Altix at NCSA (cobalt.ncsa.uiuc.edu) when
 * run on the ribosome structure:
 *   http://www.ncsa.uiuc.edu/UserInfo/Resources/Hardware/SGIAltix/
 */

#include <stdio.h>
#include <stdlib.h>
#include <cutil.h>
#include "util.h"

#define CUERR { cudaError_t err; \
  if ((err = cudaGetLastError()) != cudaSuccess) { \
  printf("CUDA error: %s, line %d\n", cudaGetErrorString(err), __LINE__); \
  return -1; }}

// max constant buffer size is 64KB, minus whatever
// the CUDA runtime and compiler are using that we don't know about
// At 16 bytes for atom, for this program 4070 atoms is about the max
// we can store in the constant buffer.
#define MAXATOMS 4070
__constant__ float4 atominfo[MAXATOMS];


// This kernel calculates simulated electron density at each grid point.
#define UNROLLX 4
__global__ static void cdensity(int numatoms, float blobbiness, float gridspacing, float * densitygrid) {
  unsigned int xindex  = (__umul24(blockIdx.x, blockDim.x) + threadIdx.x) * UNROLLX;
  unsigned int yindex  = __umul24(blockIdx.y, blockDim.y) + threadIdx.y;
  unsigned int outaddr = (__umul24(gridDim.x, blockDim.x) * UNROLLX) * yindex + xindex;

  // query current density value in the grid, start the read early
  // so the fetch occurs while we're summing the new density values
  float curdensityx1 = densitygrid[outaddr    ];
  float curdensityx2 = densitygrid[outaddr + 1];
  float curdensityx3 = densitygrid[outaddr + 2];
  float curdensityx4 = densitygrid[outaddr + 3];

  float coory = gridspacing * yindex;

  float coorx1 = gridspacing * (xindex    );
  float coorx2 = gridspacing * (xindex + 1);
  float coorx3 = gridspacing * (xindex + 2);
  float coorx4 = gridspacing * (xindex + 3);

  float densityvalx1=0.0f;
  float densityvalx2=0.0f;
  float densityvalx3=0.0f;
  float densityvalx4=0.0f;

  // Note: this implementation uses precomputed and unrolled
  // loops of dy*dy + dz*dz values for increased FP arithmetic intensity
  // per FP load.
  int atomid;
  for (atomid=0; atomid<numatoms; atomid++) {
    float dx1 = coorx1 - atominfo[atomid].x;
    float dx2 = coorx2 - atominfo[atomid].x;
    float dx3 = coorx3 - atominfo[atomid].x;
    float dx4 = coorx4 - atominfo[atomid].x;
    float dy = coory - atominfo[atomid].y;
    float dysqpdzsq = (dy * dy) + atominfo[atomid].z;
    float B_R2 = blobbiness * atominfo[atomid].w;
    densityvalx1 += expf(B_R2*(dx1*dx1 + dysqpdzsq) - blobbiness);
    densityvalx2 += expf(B_R2*(dx2*dx2 + dysqpdzsq) - blobbiness);
    densityvalx3 += expf(B_R2*(dx3*dx3 + dysqpdzsq) - blobbiness);
    densityvalx4 += expf(B_R2*(dx4*dx4 + dysqpdzsq) - blobbiness);
  }

  // accumulate density value with the existing value
  densitygrid[outaddr    ] = curdensityx1 + densityvalx1;
  densitygrid[outaddr + 1] = curdensityx2 + densityvalx2;
  densitygrid[outaddr + 2] = curdensityx3 + densityvalx3;
  densitygrid[outaddr + 3] = curdensityx4 + densityvalx4;
}


int copyatomstoconstbuf(float *atoms, int count, float zplane) {
  if (count > MAXATOMS) {
    printf("Atom count exceeds constant buffer storage capacity\n");
    return -1;
  }

  float atompre[4*MAXATOMS];
  int i;
  for (i=0; i<count*4; i+=4) {
    atompre[i    ] = atoms[i    ];
    atompre[i + 1] = atoms[i + 1];
    float dz = zplane - atoms[i + 2];
    atompre[i + 2]  = dz*dz;
    atompre[i + 3] = 1.0f / (atoms[i + 3] * atoms[i + 3]); /* 1 / r^2 */
  }

  cudaMemcpyToSymbol(atominfo, atompre, count * 4 * sizeof(float), 0);
  CUERR // check and clear any existing errors

  return 0;
}


int initatoms(float **atombuf, int count, dim3 volsize, float gridspacing) {
  dim3 size;
  int i;
  float *atoms;

  atoms = (float *) malloc(count * 4 * sizeof(float));
  *atombuf = atoms;

  // compute grid dimensions in angstroms
  size.x = gridspacing * volsize.x;
  size.y = gridspacing * volsize.y;
  size.z = gridspacing * volsize.z;

  for (i=0; i<count; i++) {
    int addr = i * 4;
    atoms[addr    ] = (rand() / (float) RAND_MAX) * size.x; 
    atoms[addr + 1] = (rand() / (float) RAND_MAX) * size.y; 
    atoms[addr + 2] = (rand() / (float) RAND_MAX) * size.z; 
    atoms[addr + 3] = ((rand() / (float) RAND_MAX) * 2.0);  // VDW radius
  }  

  return 0;
}


int main(int argc, char** argv) {
  float *doutput = NULL;
  float *density = NULL;
  float *atoms = NULL;
  dim3 volsize, Gsz, Bsz;
  rt_timerhandle runtimer, mastertimer, copytimer, hostcopytimer;
  float copytotal, runtotal, mastertotal, hostcopytotal;
  const char *statestr = "|/-\\.";
  int state=0;

  printf("CUDA accelerated electron density microbenchmark V1.0\n");
  printf("John E. Stone <johns@ks.uiuc.edu>\n");
  printf("http://www.ks.uiuc.edu/~johns/\n");
  printf("--------------------------------------------------------\n");
  int deviceCount = 0;
  cudaGetDeviceCount(&deviceCount);
  printf("Detected %d CUDA accelerators:\n", deviceCount);
  int dev;
  for (dev=0; dev < deviceCount; dev++) {
    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, dev);
    printf("  CUDA device[%d]: '%s'  Mem: %dMB  Rev: %d.%d\n", 
           dev, deviceProp.name, deviceProp.bytes / (1024*1024), 
           deviceProp.major, deviceProp.minor);
  }

  int cudadev = 0;
  if (argc == 2) {
    sscanf(argv[1], "%d", &cudadev);
    if (cudadev < 0 || cudadev >= deviceCount) {
      cudadev = 0; 
    }    
  }
  printf("  Single-threaded single-GPU test run.\n");
  printf("  Opening CUDA device %d...\n", cudadev);
  cudaSetDevice(cudadev);
  CUERR // check and clear any existing errors

  // number of atoms to simulate
  int atomcount = 1000000;

  // setup density grid size
  // XXX this is a large test case to clearly illustrate that even while
  //     the CUDA kernel is running entirely on the GPU, the CUDA runtime
  //     library is soaking up the entire host CPU for some reason.
  volsize.x = 512;
  volsize.y = 512;
  volsize.z = 1;

  // set voxel spacing
  float gridspacing = 1.5;

  // setup CUDA grid and block sizes
  // XXX we have to make a trade-off between the number of threads per
  //     block and the resulting padding size we'll end up with since
  //     each thread will do 4 consecutive grid cells in this version,
  //     we're using up some of our available parallelism to reduce overhead.
  Bsz.x =  4;                            // each thread does multiple Xs
  Bsz.y = 16;
  Bsz.z = 1;
  Gsz.x = volsize.x / (Bsz.x * UNROLLX); // each thread does multiple Xs
  Gsz.y = volsize.y / Bsz.y; 
  Gsz.z = volsize.z / Bsz.z; 

  // initialize the wall clock timers
  runtimer = rt_timer_create();
  mastertimer = rt_timer_create();
  copytimer = rt_timer_create();
  hostcopytimer = rt_timer_create();
  copytotal = 0;
  runtotal = 0;
  hostcopytotal = 0;

  printf("Grid size: %d x %d x %d\n", volsize.x, volsize.y, volsize.z);
  printf("Running kernel(atoms:%d, gridspacing %g, z %d)\n", atomcount, gridspacing, 0);

  // allocate and initialize atom coordinates and charges
  if (initatoms(&atoms, atomcount, volsize, gridspacing))
    return -1;

  // allocate and initialize the GPU output array
  int volmemsz = sizeof(float) * volsize.x * volsize.y * volsize.z;
  printf("Allocating %.2fMB of memory for output buffer...\n", volmemsz / (1024.0 * 1024.0));

  cudaMalloc((void**)&doutput, volmemsz);
  CUERR // check and clear any existing errors
  cudaMemset(doutput, 0, volmemsz);
  CUERR // check and clear any existing errors

  rt_timer_start(mastertimer);

  int iterations=0;
  int atomstart;
  for (atomstart=0; atomstart<atomcount; atomstart+=MAXATOMS) {   
    iterations++;
    int runatoms;
    int atomsremaining = atomcount - atomstart;
    if (atomsremaining > MAXATOMS)
      runatoms = MAXATOMS;
    else
      runatoms = atomsremaining;

    printf("%c\r", statestr[state]);
    fflush(stdout);
    state = (state+1) & 3;

    // copy the atoms to the GPU
    rt_timer_start(copytimer);
    if (copyatomstoconstbuf(atoms + 4*atomstart, runatoms, 0*gridspacing)) 
      return -1;
    rt_timer_stop(copytimer);
    copytotal += rt_timer_time(copytimer);
 
    // RUN the kernel...
    rt_timer_start(runtimer);
    cdensity<<<Gsz, Bsz, 0>>>(runatoms, -4.0, 0.1, doutput);
    CUERR // check and clear any existing errors
    rt_timer_stop(runtimer);
    runtotal += rt_timer_time(runtimer);
  }
  printf("Done\n");

  rt_timer_stop(mastertimer);
  mastertotal = rt_timer_time(mastertimer);

  // Copy the GPU output data back to the host and use/store it..
  density = (float *) malloc(volmemsz);
  rt_timer_start(hostcopytimer);
  cudaMemcpy(density, doutput, volmemsz,  cudaMemcpyDeviceToHost);
  CUERR // check and clear any existing errors
  rt_timer_stop(hostcopytimer);
  hostcopytotal=rt_timer_time(hostcopytimer);

#if 1
  int x, y;
  for (y=0; y<16; y++) {
    for (x=0; x<16; x++) {
      int addr = y * volsize.x + x;
      printf("out[%d]: %f\n", addr, density[addr]);
    }
  }
#endif

  printf("Final calculation required %d iterations of %d atoms\n", iterations, MAXATOMS);
  printf("Copy time: %f seconds, %f per iteration\n", copytotal, copytotal / (float) iterations);
  printf("Kernel time: %f seconds, %f per iteration\n", runtotal, runtotal / (float) iterations);
  printf("Total time: %f seconds\n", mastertotal);
  printf("Kernel invocation rate: %f iterations per second\n", iterations / mastertotal);
  printf("GPU to host copy bandwidth: %gMB/sec, %f seconds total\n",
         (volmemsz / (1024.0 * 1024.0)) / hostcopytotal, hostcopytotal);

  double atomevalssec = ((double) volsize.x * volsize.y * volsize.z * atomcount) / (mastertotal * 1000000000.0);
  printf("Efficiency metric, %g billion atom evals per second\n", atomevalssec);

  free(atoms);
  free(density);
  cudaFree(doutput);

  return 0;
}




